#![allow(dead_code)]
use ed25519_dalek::{Signer, SigningKey, VerifyingKey};
use rand::rngs::OsRng;
use std::collections::HashMap;
use std::time::Duration;
use std::time::SystemTime;
use tabled::{Table, Tabled};
use tokio::sync::Mutex;

#[derive(Clone, PartialEq)]
pub struct NodeInfo {
    pub pubkey: VerifyingKey,
    pub updated_at: SystemTime,
    pub online: bool,
}

/// Needs to be surrounded in an Arc.
pub struct Store {
    nodes: Mutex<HashMap<IP, NodeInfo>>,
    keys: Mutex<HashMap<VerifyingKey, SigningKey>>,
}
pub enum SigningError {
    CorruptedKey,
    KeyNotFound,
}

impl From<hex::FromHexError> for SigningError {
    fn from(_: hex::FromHexError) -> Self {
        Self::CorruptedKey
    }
}

impl From<ed25519_dalek::ed25519::Error> for SigningError {
    fn from(_: ed25519_dalek::ed25519::Error) -> Self {
        Self::CorruptedKey
    }
}

impl From<std::array::TryFromSliceError> for SigningError {
    fn from(_: std::array::TryFromSliceError) -> Self {
        Self::CorruptedKey
    }
}

type IP = String;

impl std::fmt::Display for SigningError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        let error_message = match self {
            SigningError::CorruptedKey => "The public key is corrupted",
            SigningError::KeyNotFound => "Did not find the public key",
        };
        write!(f, "{}", error_message)
    }
}

impl Store {
    pub fn init() -> Self {
        Self {
            nodes: Mutex::new(HashMap::new()),
            keys: Mutex::new(HashMap::new()),
        }
    }

    pub async fn add_mock_node(&self, ip: String) {
        let mut csprng = OsRng;
        let privkey = ed25519_dalek::SigningKey::generate(&mut csprng);
        self.update_node(
            ip,
            NodeInfo {
                pubkey: privkey.verifying_key(),
                updated_at: std::time::SystemTime::now(),
                online: true,
            },
        )
        .await;
        self.add_key(privkey.verifying_key(), privkey).await;
    }

    pub async fn tabled_node_list(&self) -> String {
        #[derive(Tabled)]
        struct OutputRow {
            ip: String,
            pubkey: String,
            age: u64,
        }
        let mut output = vec![];
        for (ip, node_info) in self.nodes.lock().await.iter() {
            let ip = ip.clone();
            let pubkey = hex::encode(node_info.pubkey.as_bytes());
            let age = std::time::SystemTime::now()
                .duration_since(node_info.updated_at)
                .unwrap_or(Duration::ZERO)
                .as_secs();
            output.push(OutputRow { ip, pubkey, age });
        }
        Table::new(output).to_string()
    }

    pub async fn sign_message_with_key(
        &self,
        message: &str,
        pubkey: &str,
    ) -> Result<String, SigningError> {
        let key_bytes = hex::decode(pubkey)?;
        let pubkey = VerifyingKey::from_bytes(&key_bytes.as_slice().try_into()?)?;

        let key_store = self.keys.lock().await;
        let signing_key = match { key_store.get(&pubkey) } {
            Some(k) => k,
            None => return Err(SigningError::KeyNotFound),
        };

        // TODO: check if to_bytes returns the signature in a format that people can verify from bash
        let signature = hex::encode(signing_key.sign(message.as_bytes()).to_bytes());

        Ok(signature)
    }

    pub async fn add_key(&self, pubkey: VerifyingKey, privkey: SigningKey) {
        let mut keys = self.keys.lock().await;
        keys.insert(pubkey, privkey);
    }

    pub async fn remove_key(&self, pubkey: &VerifyingKey) {
        let mut keys = self.keys.lock().await;
        keys.remove(pubkey);
    }

    async fn get_privkey(&self, pubkey: &VerifyingKey) -> Option<SigningKey> {
        let keys = self.keys.lock().await;
        keys.get(pubkey).cloned()
    }

    /// This returns true if NodeInfo got modified.
    ///
    /// On a side note, there are two types of people in this worlds:
    ///     1. Those that can extrapolate...
    pub async fn update_node(&self, ip: String, info: NodeInfo) -> bool {
        let mut nodes = self.nodes.lock().await;
        match nodes.insert(ip, info.clone()) {
            Some(old_info) => old_info.ne(&info),
            None => false,
        }
    }

    pub async fn remove_node(&self, ip: &str) {
        let mut nodes = self.nodes.lock().await;
        nodes.remove(ip);
    }

    pub async fn get_pubkey(&self, ip: &str) -> Option<NodeInfo> {
        let nodes = self.nodes.lock().await;
        nodes.get(ip).cloned()
    }

    pub async fn cycle_self_key(&self) {
        let mut csprng = OsRng;
        // TODO: save old private key to disk using SGX Sealing
        let privkey = ed25519_dalek::SigningKey::generate(&mut csprng);
        self.update_node(
            "localhost".to_string(),
            NodeInfo {
                pubkey: privkey.verifying_key(),
                updated_at: std::time::SystemTime::now(),
                online: true,
            },
        )
        .await;
        self.add_key(privkey.verifying_key(), privkey).await;
    }
}
