use crate::database;
use crate::database::NodeInfo;
use challenge::key_distribution_server::{KeyDistribution, KeyDistributionServer};
use challenge::{RemoveNodeReq, UpdateKeyReq, UpdateNodeReq};
use ed25519_dalek::SigningKey;
use prost_types::Timestamp;
use rand::rngs::OsRng;
use std::time::{Duration, SystemTime, UNIX_EPOCH};
use tonic::{transport::Server, Request, Response, Status};

pub mod challenge {
    tonic::include_proto!("challenge");
}

#[derive(Debug, Default)]
pub struct MyKeyDistribution {}

fn update_db(ip: String, privkey: String, updated_at: Option<Timestamp>) {
    let key_bytes = hex::decode(privkey).unwrap();
    let privkey = SigningKey::from_bytes(&key_bytes.as_slice().try_into().unwrap());
    let pubkey = privkey.verifying_key();
    let updated_at: std::time::SystemTime = match updated_at {
        Some(ts) => {
            let duration = Duration::new(ts.seconds as u64, ts.nanos as u32);
            UNIX_EPOCH
                .checked_add(duration)
                .unwrap_or(SystemTime::now())
        }
        None => SystemTime::now(),
    };
    database::add_node(ip.to_string(), NodeInfo { pubkey, updated_at });

    database::add_key(pubkey, privkey);
}

#[tonic::async_trait]
impl KeyDistribution for MyKeyDistribution {
    async fn update_key(&self, request: Request<UpdateKeyReq>) -> Result<Response<()>, Status> {
        let ip = request.remote_addr().unwrap().ip();
        let req = request.into_inner();
        update_db(ip.to_string(), req.keypair, req.updated_at);
        Ok(Response::new(()))
    }

    async fn update_node(&self, request: Request<UpdateNodeReq>) -> Result<Response<()>, Status> {
        let req = request.into_inner();
        update_db(req.ip, req.keypair, req.updated_at);
        Ok(Response::new(()))
    }

    async fn remove_node(&self, _request: Request<RemoveNodeReq>) -> Result<Response<()>, Status> {
        // Handle RemoveNode request
        Ok(Response::new(()))
    }
}

async fn start_client() {
    loop {
        tokio::time::sleep(Duration::from_secs(10)).await;
    }
}

async fn start_server() {
    let addr = "0.0.0.0:31373".parse().unwrap();
    let key_distribution = MyKeyDistribution::default();
    Server::builder()
        .add_service(KeyDistributionServer::new(key_distribution))
        .serve(addr)
        .await
        .unwrap();
}

pub async fn start() {
    let start_client = tokio::task::spawn(start_server());
    let start_server = tokio::task::spawn(start_server());
    futures::future::select(start_client, start_server).await;
}

pub fn add_node(ip: String) {
    let mut csprng = OsRng;
    let privkey = ed25519_dalek::SigningKey::generate(&mut csprng);
    database::add_node(
        ip,
        NodeInfo {
            pubkey: privkey.verifying_key(),
            updated_at: std::time::SystemTime::now(),
        },
    );
    database::add_key(privkey.verifying_key(), privkey);
}
